/*
** SpeciesSet.h
*/

# ifndef SPECIESSET_H
# define SPECIESSET_H

/* OVERVIEW:
 * SpeciesSets are unbounded, mutable sets of Species
 */

# include "bool.h"
# include "Species.h"

# include "SpeciesSet.rh"

extern /*@only@*/ SpeciesSet SpeciesSet_new (void) 
   /* EFFECTS:  Allocates and returns a new empty SpeciesSet; i.e. { } */ ;

extern void SpeciesSet_free (/*@only@*/ SpeciesSet p_set)
   /*@modifies p_set@*/ 
   /* MODIFIES: set
   ** EFFECTS:  Destroys set and releases memory occupied by it
   */ ;

extern void SpeciesSet_insert (SpeciesSet p_set,
                               /*@dependent@*/ Species p_species)
   /*@modifies p_set@*/ 
   /* MODIFIES: set
   ** EFFECTS: Adds species to elements of set: 
   **	set_post = set_pre U { species }
   */ ;

extern void SpeciesSet_remove (SpeciesSet p_set,
			       Species p_species)
   /*@modifies p_set@*/ 
   /* MODIFIES: set
   ** EFFECTS: Removes species from set, if present:
   **	set_post = set_pre - { species }
   */ ;

extern bool SpeciesSet_contains (SpeciesSet p_set,
				/*@notnull@*/ const char *p_name)
   /* EFFECTS: Returns true iff set contains a species named name.
   */ ;

extern SpeciesSet SpeciesSet_copy (SpeciesSet p_set)
   /* EFFECTS: Allocates and returns a copy of set */ ;

extern void SpeciesSet_split (SpeciesSet p_orig, unsigned p_bits,
			      SpeciesSet p_left, SpeciesSet p_right)
   /*@modifies p_left, p_right@*/ 
   /* REQUIRES: left = { } AND right = { }
   ** MODIFIES: left, right
   ** EFFECTS:  Splits orig into two subsets, left and right, such that:
   **		orig = left U right
   **		intersect(left,right) = { }
   */ ;

extern unsigned SpeciesSet_size (SpeciesSet p_set) 
   /* EFFECTS: Returns the number of elements in set */ ;

extern bool SpeciesSet_isEmpty (SpeciesSet p_set)
  /* EFFECTS: Returns true iff the set has no elements */ ;

/* SpeciesSet_forall is an iterator that loops over all elements in a 
 * SpeciesSet, and performs a specified action on each element.
 *
 * It is used as follows:
 *
 *	SpeciesSet set;
 *	...
 *	SpeciesSet_forall(set,element) {
 *	    // loop body using <element>
 *	} end_SpeciesSet_forall;
 *
 * The macro writes a for loop that iterates over all elements of the set.
 * It declares a temporary variable <element>, of type Species, and uses it
 * as the index variable of the loop.
 *
 * This is implemented as a macro because C does not have the equivalent of 
 * Java Enumerations or C++ Iterators.  The Java equivalent would be:
 *
 *	Enumeration e = set.elements();
 *	while (e.hasMoreElements()) {
 *	    Species element = (Species) e.getNextElement();
 *	    // loop body using <element>
 *	}
 *
 * You do not need to understand how the macro works in order to complete 
 * this assignment, nor do you need to add any additional calls to it.
 */

/*@iter SpeciesSet_forall (SpeciesSet s, yield dependent Species el) @*/

# include "SpeciesSet.ih"

#endif
